<?php

namespace Migo\Clients;

use log as log, Migo;
use cache as cache;
use Exception as Exception;


//const VT_BASE_URI = "https://app.migo-thermostat.com/";
const VT_BASE_URI = "https://api.netatmo.net/";
const VT_SERVICES_URI = "https://api.netatmo.net/api";
const VT_ACCESS_TOKEN_URI = "https://api.netatmo.net/oauth2/token";
const VT_AUTHORIZE_URI = "https://api.netatmo.net/oauth2/authorize";
const CL_LOG= "migoThermostat";
/**
 * OAuth2.0 Netatmo client-side implementation.
 *
 * @author Originally written by Thomas Rosenblatt <thomas.rosenblatt@netatmo.com>.
 */
class VTApiClient{
    /**
   * Array of persistent variables stored.
   */
    protected $conf = array();
    protected $refresh_token;
    protected $access_token;
    protected $expires_at;
/* ****************************************************************  */
	public function getVariable($name, $default = NULL){
        return isset($this->conf[$name]) ? $this->conf[$name] : $default;
    }
/* ****************************************************************  */
    public function getRefreshToken(){
        return $this->refresh_token;
    }

    /* ****************************************************************  */
    public function setVariable($name, $value){
        $this->conf[$name] = $value;
        return $this;
    }
/* ****************************************************************  */
    private function updateSession($value){
        $cb = $this->getVariable("func_cb");
      	$object = $this->getVariable("object_cb");
      	$return = false;
      //$object = new parent();
        if($object && $cb){
            if(method_exists($object, $cb)){
              	$return = call_user_func_array([ $object, $cb], [$value]);
            }
        }
        else if($cb && is_callable($cb)){
            $return = call_user_func_array($cb, [$value]);
        }
      	return $return;
    }
/* ****************************************************************  */
    private function setTokens($value){
        if(isset($value["access_token"])){
            $this->access_token = $value["access_token"];
            $update = true;
        }
        if(isset($value["refresh_token"])){
            $this->refresh_token = $value["refresh_token"];
            $update = true;
        }
        if(isset($value["expires_in"])){
            $this->expires_at = time() + $value["expires_in"] - 30;
        }
      	cache::set('vtApi_ArToken', $value);
        if(isset($update)) $this->updateSession($value);
    }

/* ****************************************************************  */
    public function setTokensFromStore($value){
        if(isset($value["access_token"]))
            $this->access_token = $value["access_token"];
        if(isset($value["refresh_token"]))
            $this->refresh_token = $value["refresh_token"];
      	cache::set('vtApi_ArToken', $value);
    }
/* ****************************************************************  */
    public function unsetTokens(){
        $this->access_token = null;
        $this->refresh_token = null;
        $this->expires_at = null;
      	cache::set('vtApi_ArToken', []);
    }
/* ****************************************************************  */
    public function __construct($config = array()){
        // If tokens are provided let's store it
        if(isset($config["access_token"])){
            $this->access_token = $config["access_token"];
            unset($access_token);
        }
      	if(isset($config["refresh_token"])){
            $this->refresh_token = $config["refresh_token"];
        }
      	// We must set uri first.
        $uri = [
          			"base_uri" => VT_BASE_URI, 
          			"services_uri" => VT_SERVICES_URI, 
          			"access_token_uri" => VT_ACCESS_TOKEN_URI, 
          			"authorize_uri" => VT_AUTHORIZE_URI
        ];
        foreach($uri as $key => $val){
            if(isset($config[$key])){
                $this->setVariable($key, $config[$key]);
                unset($config[$key]);
            }
            else{
                $this->setVariable($key, $val);
            }
        }

        if(isset($config['scope']) && is_array($config['scope'])){
            foreach($config['scope'] as $scope){
                trim($scope);
            }

            $scope = implode(' ', $config['scope']);
            $this->setVariable('scope', $scope);
            unset($config['scope']);
        }

        // Other else configurations.
        foreach ($config as $name => $value){
            $this->setVariable($name, $value);
        }

        if($this->getVariable("code") == null && isset($_GET["code"])){
            $this->setVariable("code", $_GET["code"]);
        }
  }
/* ****************************************************************  */
     private static function normalizeArray($array) {
        foreach ($array as &$el) {
            if (is_bool($el)) {
              $el = ($el) ? "true" : "false";
            } elseif (is_array($el)) {
              $el = self::normalizeArray($el);
            }
        }
        return $array;
    }
/* ****************************************************************  */
    private static function formatArray($array){
      	//log::add(CL_LOG, 'debug', __FUNCTION__ .' in array: ' . json_encode($array, JSON_UNESCAPED_UNICODE,JSON_PRETTY_PRINT));
		$result=[];
        foreach ($array as $key => $value){
			if(is_array($value)){
              	if(is_numeric($key)) $key = json_encode($key);
              	$result[$key] = self::formatArray($value);
            }
          	//elseif (!is_string($value) && !is_array($value)){
			elseif (is_bool($value) || $value == null){
				$result[$key] = json_encode($value);//urldecode($NameValueParts['value']);
			}else{
              	$result[$key] = $value;
            }
		}
      	//log::add(CL_LOG, 'debug', __FUNCTION__ .' result: ' . json_encode($result, JSON_UNESCAPED_UNICODE,JSON_PRETTY_PRINT));
        
      	return $result;
      
    }
/* ****************************************************************  */
    private function makeRequest($_path, $method = 'GET', $params=null, $cookie = false, $headers = array(), $retry = false){
		$method = strtoupper($method);
      	$opts =  [
            CURLOPT_CONNECTTIMEOUT => 10,
            CURLOPT_RETURNTRANSFER => TRUE,
            CURLOPT_HEADER         => TRUE,
            CURLOPT_TIMEOUT        => 80,
           	CURLOPT_USERAGENT      => "migoThermostat(Jeedom)",
        	CURLOPT_SSL_VERIFYPEER => TRUE,
          	CURLINFO_HEADER_OUT	   => TRUE,
          	CURLOPT_FOLLOWLOCATION => TRUE,
           // CURLOPT_HTTPHEADER     => array("Accept: application/json", 'Expect:', 'Content-Type: application/json','Connection: Keep-Alive'),
        ];
      	$generalHeaders = [
					"Accept" => "application/json",
					"Expect" => "",
					"Content-Type" => "application/json",
					//"Connection" => "Keep-Alive",
        ];
      	//override genral header
      	$headers = array_merge($generalHeaders, $headers);
        
          //batch request or conform it:
		if($cookie){
          	$cookiefile = jeedom::getTmpFolder(CL_LOG).'/cookie_migoThermostat';
          	$opts[CURLOPT_COOKIEJAR] =  $cookiefile;
          	$opts[CURLOPT_COOKIEFILE] = $cookiefile ;
        }
		if( is_array($params) && !empty($params) ){
			if(isset($params['access_token'])){
                $headers[] = 'Authorization: Bearer ' . $params['access_token'];
                unset($params['access_token']);
            }
          	if(isset($headers['access_token'])){
                $headers[] = 'Authorization: Bearer ' . $headers['access_token'];
                unset($headers['access_token']);
            }
          	//$params = self::formatArray($params);
          	
            //$query = http_build_query($params);
        	//$lenght = strlen($query);
		}
      	if($method == 'GET'){
          	$query = "";
            $opts[CURLOPT_HTTPGET] = TRUE;
          	if (!is_array($params)){
				if( json_decode($params)){
                  	$query = http_build_query(json_decode($params));
                }elseif( $params != "") {
                  	$query = $params;
                }
              	//log::add(CL_LOG, 'debug', __FUNCTION__ ." $method Not array query: ".$query);
          	}
			else{//is array
              	$params = self::formatArray($params);
				$query = http_build_query($params);
              	//log::add(CL_LOG, 'debug', __FUNCTION__ ." $method array query: ".$query);
              	
          	}
          	$path = $_path."?".$query;
            $payload=$query;
          	//$lenght = strlen($query);
          	//log::add(CL_LOG, 'debug', __FUNCTION__ ." $method query: ".$query);
        }
      	else{//if($method == 'POST')
          	if(is_array($params)){//is array
				$payload = json_encode($params);
				//log::add(CL_LOG, 'debug', __FUNCTION__ ." $method array: ".$payload);
			}
          	elseif(json_decode($params)){
                $payload = $params;
              	//log::add(CL_LOG, 'debug', __FUNCTION__ ." $method json: ".$payload);
            }
          	else{
				$headers['Content-Type'] = "application/x-www-form-urlencoded";
                $payload = $params;
				//log::add(CL_LOG, 'debug', __FUNCTION__ ." $method string: ".$payload);
			}
			$path = $_path;
          	$opts[CURLOPT_CUSTOMREQUEST] = $method;
          	//$opts[CURLOPT_POST] = true;
          	$opts[CURLOPT_POSTFIELDS] = $payload;
          	$lenght = strlen($payload);
          	$lenght ? $headers['Content-Length'] = $lenght : "" ;
      	
          
        }
      	
      	//log::add(CL_LOG, 'debug', __FUNCTION__ ." $method ".parse_url($_path, PHP_URL_PATH)." ".$payload );
      
      
		foreach($headers as $key=>$value){
            if(!is_numeric($key)){
            	$opts[CURLOPT_HTTPHEADER][] = $key.': '.$value;
            }else{
            	$opts[CURLOPT_HTTPHEADER][] = $value;
            }
		}
        //log::add(CL_LOG, 'debug', __FUNCTION__ ." HTTPHEADER: ".var_export($opts[CURLOPT_HTTPHEADER], true));
      	$opts[CURLOPT_URL] = $path;
      	$ch = curl_init();
        curl_setopt_array($ch, $opts);
      
		$result = curl_exec($ch);
      	$errno = curl_errno($ch);
        // CURLE_SSL_CACERT || CURLE_SSL_CACERT_BADFILE
        if ($errno == 60 || $errno == 77){
            log::add(CL_LOG, 'error', __FUNCTION__ ." WARNING ! SSL_VERIFICATION has been disabled...\n");
          	curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, FALSE);
            $result = curl_exec($ch);
        }
      	
      	//log::add(CL_LOG, 'debug', __FUNCTION__ .' result: ' . $result);
        $http_code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
		//log::add(CL_LOG, 'debug', __FUNCTION__ .' http_code: '.$http_code);
      	$header_size = curl_getinfo($ch, CURLINFO_HEADER_SIZE);
      	$request_header = curl_getinfo($ch,CURLINFO_HEADER_OUT);//need CURLOPT_HEADER
      	//log::add(CL_LOG, 'debug', __FUNCTION__ .' request_header: ' . $request_header);
        
      
      	if ($result === FALSE){
          	$er_code = curl_errno($ch);
          	$er_msg = ' curl Error ('.curl_errno($ch).") : ".curl_error($ch);
          	//curl_close($ch);
          	if ($er_code == 28 && $retry == false){
              	log::add(CL_LOG, 'warning', __FUNCTION__ .' er_code: ' . $er_code. " ... retry" );
              	sleep(2);
              	return $this->makeRequest($_path, $method , $params, $cookie, $headers, true);
            }
            
          	throw new Exception($er_msg, $er_code);
        }
      	curl_close($ch);
		if( isset($opts[CURLOPT_HEADER]) && $opts[CURLOPT_HEADER]==true ){
          	$rawHeaders = explode("\r\n", substr($result, 0, $header_size));
          	$body = substr($result, $header_size);
          	//log::add(CL_LOG, 'debug', __FUNCTION__ .' rawHeaders: '.json_encode($rawHeaders, JSON_UNESCAPED_UNICODE,JSON_PRETTY_PRINT));
          	//log::add(CL_LOG, 'debug', __FUNCTION__ .' body: ' . $body);
          	preg_match('/^HTTP\/1.1 ([0-9]{3,3}) (.*)$/', $rawHeaders[0], $matches);
			$headerCode = isset($matches[1]) ? $matches[1] : 0;
			$headerStatus = isset($matches[2]) ? strtolower($matches[2]) : "Nok";
			//log::add(CL_LOG, 'debug', __FUNCTION__ .' headerCode: '.$headerCode.'-- headerStatus: '.$headerStatus);
		}else{
          	$rawHeaders = explode("\r\n", $request_header);
          	$body = $result;
        }
      	if(substr($http_code, 0, 1) === '2'){
            $decode = json_decode($body, TRUE);
            if(!$decode){
                $rsp_body = $body;
              	//log::add(CL_LOG, 'debug', __FUNCTION__ .' !decode: '.$rsp_body);
            } 
          	elseif(isset($decode['body'])){
              	//log::add(CL_LOG, 'error', __FUNCTION__ .' 200 decode && body: '.json_encode($decode, true));
            	if(isset($decode['body']['errors'])){
                  	foreach ($decode['body']['errors'] as $error) {
                      	if(isset($error['code'])){
                          	$errmsg = ( isset($error['command']) ) ? $error['command'] .': '.self::$err200[$error['code']] : self::$err200[$error['code']];
                          	$errors[$error['code']] = $errmsg;
                        }
                    }
                  	$decode['body']['errors'] = $errors;
                } 
				$rsp_body = $decode['body'];
				$rsp_status = strtolower($decode['status']);
            }
          	else{
				return $decode;
                //log::add(CL_LOG, 'error', __FUNCTION__ .' 200 decode && !body: '.json_encode($decode, true));
            }
            $return['code']= $http_code;
          	$return['body'] = $rsp_body; 
          	$return['status']= $rsp_status ? $rsp_status : $headerStatus;
          	$return['headers'] = $rawHeaders;
          	
        }
        else{//http_code != 200
          	//log::add(CL_LOG, 'warning', __FUNCTION__ .'  ! http_code: '.$http_code.' at '.parse_url($_path, PHP_URL_PATH));
				
            $decode = json_decode($body, TRUE);
			if(!$decode){
              	$code= $http_code ? $http_code : $headerCode ;
              	$code= $code ? $code : $headerStatus ; 
              	
				if( $http_code != 200 ){
					$err_msg = " Request error at ".parse_url($_path, PHP_URL_PATH)." : " .$headerStatus;
                  	throw new Exception($err_msg, $code);
				}elseif( is_string($body) ){
                  	$err_msg = " Request error at ".parse_url($_path, PHP_URL_PATH)." : " .$body;
					throw new Exception($err_msg, $code);
                }
              	log::add(CL_LOG, 'warning', __FUNCTION__ .' !decode: '.$code.$err_msg);
              	
            }else{
               	$status = array_keys($decode)[0];
               	//log::add(CL_LOG, 'debug', __FUNCTION__ .' status '.$status);
				$msg = array_values($decode)[0];
              	if($status == "error" || isset($decode["error"])){
                //log::add(CL_LOG, 'debug', __FUNCTION__ .' msg: '.$msg['message']);
               		//$apiEr_msg = ( isset($msg['message']) )?$msg['message'] : isset($decode['error_description']) ? $decode['error_description']: "";
                	if(isset($msg['message'])){
                      $apiEr_msg = $msg['message'];
                    }elseif(isset($decode['error_description'])){
                      $apiEr_msg = $decode['error_description'];
                    }elseif(is_string($decode['error'])){
                      $apiEr_msg = $decode['error'];
                    }else{
                       $apiEr_msg =' unknown';
                    }
                  	$apiEr_code = (isset($msg['code']))  ? $msg['code']: $http_code;
                  	$ermsg = " API_Erreur at ".parse_url($_path, PHP_URL_PATH). " : ".$apiEr_msg."($apiEr_code)" ;//(code $apiEr_code) 
                  	log::add(CL_LOG, 'warning', __FUNCTION__ .' '.$ermsg);
                  	$err['error']['code'] = $apiEr_code;
                    $err['error']['message'] = $apiEr_msg;
                  	//throw new ExceptionException($ermsg, null);
                    throw new Exception($ermsg, $apiEr_code);
                }
              	//log::add(CL_LOG, 'debug', __FUNCTION__ .' '.$ermsg);
            }
          	$return['code']= ($apiEr_code !== null) ? $apiEr_code : $http_code;
          	$return['body'] = $apiEr_msg;
          	$return['status'] = strtolower($status);//strtolower($headerStatus);
          	
        }
      	return $return;
      
      
	}
/* ****************************************************************  */
    public function getAccessToken(){
        //find best way to retrieve access_token
        if($this->access_token){
          	if(!is_null($this->expires_at) && $this->expires_at < time()){ // access_token expired.
            	if($this->refresh_token){
                    return $this->getAccessTokenFromRefreshToken($this->refresh_token);//exception will be thrown otherwise
                }
                else{
                    throw new Exception("Access token expired");
                }
            }
            return array("access_token" => $this->access_token);
        }
      	if($this->getVariable('code')){// grant_type == authorization_code.
        	return $this->getAccessTokenFromAuthorizationCode($this->getVariable('code'));
        }
      	else if($this->refresh_token){// grant_type == refresh_token
        	return $this->getAccessTokenFromRefreshToken();
        }
        else if($this->getVariable('username') && $this->getVariable('password') && $this->getVariable('user_prefix')){
        	return $this->getAccessTokenFromPassword($this->getVariable('username'), $this->getVariable('password'));
        }
        else throw new Exception("No access token stored");
    }
  
/* ****************************************************************  */
    public function revokeToken($client_id, $client_secret){
        $revokeUri = 'https://api.netatmo.com/oauth2/revoke';
    	
    	$headers=['Content-Type' => "application/x-www-form-urlencoded"];
        if ($token = $this->refresh_token != NUL){
            $token_type = "refresh_token";
        }
    	elseif ($token = $this->access_token != NUL){
            $token_type = "access_token";
        }
    	
    
        if (($client_id = $this->getVariable('client_id')) != NULL && ($client_secret = $this->getVariable('client_secret')) != NULL){
			$params = [
          			'client_id' => $client_id,
                    'client_secret' => $client_secret,
                    'app_version" => "1.4.0.1', 
                    'token_type_hint' => $token_type,//"refresh_token",
                    'token' => $token,
        	];
          	$ret = $this->makeRequest($revokeUri,'POST',$params,false,$headers );
			return $ret;
		}
            else
			throw new Exception("missing args for disconnect ");
        }
/* ****************************************************************  */
    public function getAuthorizeUrl($state = null){
        $redirect_uri = $this->getRedirectUri();
        if($state == null){
            $state = rand();
        }
        $scope = $this->getVariable('scope');
        $params = array(
          				"app_version" => "1.4.0.1", 
                        "scope" => $scope, 
                        "state" => $state, 
                        "client_id" => $this->getVariable("client_id"), 
                        "client_secret" => $this->getVariable("client_secret"), 
                        "response_type" => "code", 
                        "redirect_uri" => $redirect_uri);
        return $this->getUri($this->getVariable("authorize_uri"), 'GET',$params);
    }

/* ****************************************************************  */
    private function getAccessTokenFromAuthorizationCode($code){
        $redirect_uri = $this->getRedirectUri();
        $scope = $this->getVariable('scope');
        if($this->getVariable('access_token_uri') && ($client_id = $this->getVariable('client_id')) != NULL 
           && ($client_secret = $this->getVariable('client_secret')) != NULL && $redirect_uri != NULL){
            $params=[
                    'grant_type' => 'authorization_code',
                    'client_id' => $client_id,
                    'client_secret' => $client_secret,
                    'code' => $code,
                    'redirect_uri' => $redirect_uri,
                    'scope' => $scope,
            ];
          	$ret = $this->makeRequest($token_uri,'POST',http_build_query($params, NULL, '&'));
          	$this->setTokens($ret);
            return $ret;
        }
        else
            throw new Exception("missing args for getting authorization code grant");
    }

/* ****************************************************************  */
    public function getAccessTokenFromPassword($username, $password){
        $scope = $this->getVariable('scope');
		$token_uri=$this->getVariable('access_token_uri');
      	if ($this->getVariable('access_token_uri') 
			&& ($client_id = $this->getVariable('client_id')) != NULL 
		    && ($client_secret = $this->getVariable('client_secret')) != NULL
		    && ($user_prefix = $this->getVariable('user_prefix')) != NULL){
            $params=[
                    'username' => $username,
                    'password' => $password,
                    'client_id' => $client_id,
                    'client_secret' => $client_secret,
                    'grant_type' => 'password',//'password',
                    'scope' => $scope,//'read_thermostat write_thermostat',//??ESPACE
					'user_prefix' => $user_prefix,
            ];
          	$ret = $this->makeRequest($token_uri,'POST',http_build_query($params, NULL, '&'));
          	
          	$this->setTokens($ret);
            return $ret;
        }
        else
            throw new Exception("missing args for getting password grant");
    }
/* ****************************************************************  */
    private function getAccessTokenFromRefreshToken(){
        if ($this->getVariable('access_token_uri') && ($client_id = $this->getVariable('client_id')) != NULL 
            && ($client_secret = $this->getVariable('client_secret')) != NULL && ($refresh_token = $this->refresh_token) != NULL){
            if($this->getVariable('scope') != null){
                $params=[
                    	'grant_type' => 'refresh_token',
                        'client_id' => $this->getVariable('client_id'),
                        'client_secret' => $this->getVariable('client_secret'),
                        'refresh_token' => $refresh_token,
                        'scope' => $this->getVariable('scope'),
                  		'user_prefix' => $user_prefix,
                ];
          		$ret = $this->makeRequest($token_uri,'POST',http_build_query($params, NULL, '&'));
          
            }
            else{
                $params=[
                    	'grant_type' => 'refresh_token',
                        'client_id' => $this->getVariable('client_id'),
                        'client_secret' => $this->getVariable('client_secret'),
                        'refresh_token' => $refresh_token,
                  		'user_prefix' => $user_prefix,
                  	
            	];
				$ret = $this->makeRequest($token_uri,'POST',http_build_query($params, NULL, '&'));
            }
            $this->setTokens($ret);
            return $ret;
        }
        else
            throw new Exception("missing args for getting refresh token grant");
    }
/* ****************************************************************  */
    protected function makeOAuth2Request($path, $method = 'GET', $params = array(), $reget_token = true){
        try{
            $res = $this->getAccessToken();
        }
        catch(Exception $ex){
            throw new Exception($ex->getMessage(), $ex->getCode());
        }
        if(is_array($params)) $params["access_token"] = $res["access_token"];
        try{
            $res = $this->makeRequest($path, $method, $params);
            return $res;
        }
        catch(Exception $ex) {
          	
            if($reget_token == true){
                //log::add(CL_LOG, 'info', '  ******'.__FUNCTION__ .' 860 reget_token '.$ex->getCode());
            
              	switch($ex->getCode()){
                  case 2://VTRestErrorCode::INVALID_ACCESS_TOKEN:
                  case 3:// VTRestErrorCode::ACCESS_TOKEN_EXPIRED:
                    	$oldtoken = $this->access_token;
                        $this->access_token = null;
                        //log::add(CL_LOG, 'warning', '  ****** '.__FUNCTION__ .' error '.self::$API_ERR[$ex->getCode()]);
                    	//Ok token has expired let's retry once
                    	$reniew = $this->updateSession(self::$API_ERR[$ex->getCode()]);
                    	if($oldtoken != $reniew){
                          	$this->access_token = $reniew;
                          	$params["access_token"] = $reniew;
                          	$res = $this->makeRequest($path, $method, $params);
                    		return $res;
                        }
                    	elseif($this->refresh_token){
                          	log::add(CL_LOG, 'warning', '  ******'.__FUNCTION__ .' 868 is_refresh_token ');
                            try{
                                $this->getAccessTokenFromRefreshToken();//exception will be thrown otherwise
                            }
                            catch(Exception $ex2){
                                //Invalid refresh token TODO: Throw a special exception
                                log::add(CL_LOG, 'warning', '  ******'.__FUNCTION__ .' 875 Exception '.$ex2->getMessage());
                    			throw $ex;
                            }
                        }
                    	else if( $this->getVariable('client_secret') ){
                            log::add(CL_LOG, 'warning', '  ******'.__FUNCTION__ .' 880 is_client_secret ');
                            try{
                                $this->getAccessToken();//exception will be thrown otherwise
                              
                            }
                            catch(Exception $ex2){
                                //Invalid refresh token TODO: Throw a special exception
                                log::add(CL_LOG, 'warning', '  ******'.__FUNCTION__ .' 887 Exception '.$ex2->getMessage());
                    			throw $ex;
                            }
                        }
                        else{
                           log::add(CL_LOG, 'warning', '  ******'.__FUNCTION__ .' 893 Exception '.$ex->getMessage());
                    			
                          throw $ex;
                        }
                        return $this->makeOAuth2Request($path, $method, $params, false);
                    break;
                    case 13:
                    	//log::add(CL_LOG, 'debug', '  ******'.__FUNCTION__ .' 891 default Exception '.$ex->getMessage());
                 		throw $ex;
                    	return null;
                    break;
                    default:
                    	//log::add(CL_LOG, 'info', '  ******'.__FUNCTION__ .' 895 default Exception '.$ex->getMessage());
                 		throw $ex;
                }
            }
            else{//$reget_token
                 throw $ex;
           	}
        }
        return $res;
    }
/* ****************************************************************  */
    public function api($path, $method = 'GET', $params = array(), $secure = false){
        if (is_array($method) && empty($params)){
            $params = $method;
            $method = 'GET';
        }

        $res = $this->makeOAuth2Request($this->getUri($path, array(), $secure), $method, $params);
        if(isset($res["body"])) return $res["body"];
    	else return $res;
    }
/* ****************************************************************  */
    public function noTokenApi($path, $method = 'GET', $params = array()){
        if (is_array($method) && empty($params)){
            $params = $method;
            $method = 'GET';
        }

        // json_encode all params values that are not strings.
        foreach ($params as $key => $value){
            if (!is_string($value)){
                $params[$key] = json_encode($value);
            }
        }

        return $this->makeRequest($path, $method, $params);
    }
/* ****************************************************************  */
    public static function str_replace_once($str_pattern, $str_replacement, $string){
        if (strpos($string, $str_pattern) !== false){
            $occurrence = strpos($string, $str_pattern);
            return substr_replace($string, $str_replacement, strpos($string, $str_pattern), strlen($str_pattern));
        }
        return $string;
    }
	
/* ****************************************************************  */
    private function getRequestUri(){
        if (isset($_SERVER['REQUEST_URI'])) {
            $uri = $_SERVER['REQUEST_URI'];
        }
        else {
            if (isset($_SERVER['argv'])) {
                $uri = $_SERVER['SCRIPT_NAME'] . '?' . $_SERVER['argv'][0];
            }
            elseif (isset($_SERVER['QUERY_STRING'])) {
                $uri = $_SERVER['SCRIPT_NAME'] . '?' . $_SERVER['QUERY_STRING'];
            }
            else {
                $uri = $_SERVER['SCRIPT_NAME'];
            }
        }
        // Prevent multiple slashes to avoid cross site requests via the Form API.
        $uri = '/' . ltrim($uri, '/');

        return $uri;
    }
/* ****************************************************************  */
    protected function getCurrentUri() {
        $protocol = isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == 'on'
          ? 'https://'
          : 'http://';
        $current_uri = $protocol . $_SERVER['HTTP_HOST'] . $this->getRequestUri();
        $parts = parse_url($current_uri);

        $query = '';
        if (!empty($parts['query'])) {
          $params = [];
          parse_str($parts['query'], $params);
          $params = array_filter($params);
          if (!empty($params)) {
            $query = '?' . http_build_query($params, NULL, '&');
          }
        }

        // Use port if non default.
        $port = isset($parts['port']) &&
          (($protocol === 'http://' && $parts['port'] !== 80) || ($protocol === 'https://' && $parts['port'] !== 443))
          ? ':' . $parts['port'] : '';

        // Rebuild.
        return $protocol . $parts['host'] . $port . $parts['path'] . $query;
    }

/* ****************************************************************  */
    protected function getRedirectUri(){
        $redirect_uri = $this->getVariable("redirect_uri");
        if(!empty($redirect_uri)) return $redirect_uri;
        else return $this->getCurrentUri();
    }

/* ****************************************************************  */
    protected function getUri($path = '', $params = array(), $secure = false){
        $url = $this->getVariable('services_uri') ? $this->getVariable('services_uri') : $this->getVariable('base_uri');
        if($secure == true){
            $url = self::str_replace_once("http", "https", $url);
        }
        if(!empty($path))
            if (substr($path, 0, 4) == "http")
                $url = $path;
            else if(substr($path, 0, 5) == "https")
                $url = $path;
            else
                $url = rtrim($url, '/') . '/' . ltrim($path, '/');

        if (!empty($params))
            $url .= '?' . http_build_query($params, NULL, '&');

        return $url;
    }
/* ****************************************************************  */
    public function getPartnerDevices(){
        return $this->api("partnerdevices", "POST");
    }
/* ****************************************************************  */
    protected function addWebhook($url, $app_type) {
        $params = ['url' => $url, 'app_type' => $app_type];
        $this->api('addwebhook', $params);
    }

/* ****************************************************************  */
    protected function dropWebhook($app_type){
        $params = ['app_type' => $app_type];
        $this->api('dropwebhook', $params);
    }
/* ****************************************************************  */
    public static $API_ERR = [
                        1=>'Access token missing',
                        2=>'Invalid access token',
                        3=>'Access token expired',
                        4=>'Inconsistency error',
                        5=>'Application deactivated',
                        6=>'Invalid email',
                        7=>'Nothing to modify',
                        8=>'Email already exists',
                        9=>'Device not found',
                        10=>'Missing args',
                        11=>'Internal error',
                        12=>'Device or secret no match',
                        13=>'Operation forbidden',
                        14=>'Application name already exists',
                        15=>'No places in device',
                        16=>'Mgt key missing',
                        17=>'Bad mgt key',
                        18=>'Device id already exists',
                        19=>'Ip not found',
                        20=>'Too many user with ip',
                        21=>'Invalid arg',
                        22=>'Application not found',
                        23=>'User not found',
                        24=>'Invalid timezone',
                        25=>'Invalid date',
                        26=>'Max usage reached',
                        27=>'Measure already exists',
                        28=>'Already device owner',
                        29=>'Invalid ip',
                        30=>'Invalid refresh token',
                        31=>'Not found',
                        32=>'Bad password',
                        33=>'Force associate',
                        34=>'Module already paired',
                        35=>'Unable to execute',
                        36=>'Prohibitted string',
                        37=>'Camera no space available',
                        38=>'Password complexity too low',
                        39=>'Too many connection failure',
];
/* ****************************************************************  */
    public static $HTTPCODE = [
        100 => 'Continue',
        101 => 'Switching Protocols',
        200 => 'OK',
        201 => 'Created',
        202 => 'Accepted',
        203 => 'Non-Authoritative Information',
        204 => 'No Content',
        205 => 'Reset Content',
        206 => 'Partial Content',
        300 => 'Multiple Choices',
        301 => 'Moved Permanently',
        302 => 'Found',
        303 => 'See Other',
        304 => 'Not Modified',
        305 => 'Use Proxy',
        306 => '(Unused)',
        307 => 'Temporary Redirect',
        400 => 'Bad Request',
        401 => 'Unauthorized',
        402 => 'Payment Required',
        403 => 'Forbidden',
        404 => 'Not Found',
        405 => 'Method Not Allowed',
        406 => 'Not Acceptable',
        407 => 'Proxy Authentication Required',
        408 => 'Request Timeout',
        409 => 'Conflict',
        410 => 'Gone',
        411 => 'Length Required',
        412 => 'Precondition Failed',
        413 => 'Request Entity Too Large',
        414 => 'Request-URI Too Long',
        415 => 'Unsupported Media Type',
        416 => 'Requested Range Not Satisfiable',
        417 => 'Expectation Failed',
        500 => 'Internal Server Error',
        501 => 'Not Implemented',
        502 => 'Bad Gateway',
        503 => 'Service Unavailable',
        504 => 'Gateway Timeout',
        505 => 'HTTP Version Not Supported',
	];
/* ****************************************************************  */
    public static $err200 = [
                  '1' => 'unknown_error',
                  '2' => 'internal_error',
                  '3' => 'parser_error',
                  '5' => 'command_invalid_params',
                  '6' => 'device_unreachable',
                  '7' => 'command_error',
                  '8' => 'battery_level',
                  '14' => 'busy',
                  '19' => 'module_unreachable',
                  '23' => 'nothing_to_modify',
                  '27' => 'temporarily_banned'
	];



}




?>